/* ----------------------------------------------------------------------------
 * hook_input version 0.1
 * Hook input event.
 *
 * Copyright 2004 Takuo Yonezawa (takuo@aya.or.jp)
 *
 * A large amount of code borrowed from mouseemu
 * Copyright 2003, 2004 Colin Leroy (colin@colino.net).
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * ----------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <linux/input.h>
#include "linux/uinput.h"

static int running = -1;
static int ui_fd = -1;

extern void handle_event(int fd, struct input_event event, void *user_data);
extern void set_event_bits(int fd);
extern void *init_event_handler(int number_of_event_device_files, int event_fds[]);
extern void finalize_event_hander(void *user_data);
extern char *name;
extern char *version_information;
extern char *description;

static void print_usage(FILE *output) 
{
  fprintf(output,
          "Usage: %s [options] device [devices...]\n"
          "\n"
          "%s"
          "\n"
          "Options:\n"
          "  -h, --help: print this message\n"
          "  -v, --version: print version information\n"
          "  -u file_name, --uinput_device file_name: set uinput device name\n",
          name, description);
}

static void print_version(void)
{
  printf("%s\n"
         "based on hook_input version 0.1\n"
         "\n"
         "Copyright (C) 2004 Takuo Yonezawa (takuo@aya.or.jp)\n"
         "A large amount of code borrowed from mouseemu\n"
         "Copyright 2003, 2004 Colin Leroy (colin@colino.net).\n"
         "\n"
         "This is free software; see the source for copying conditions.\n"
         "There is NO warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n",
         version_information);
}

static void term_handler(int sig)
{
  running = -sig;
}

static void install_sighandler(void)
{
  sigset_t mask;

  sigemptyset(&mask);

#ifdef SIGTERM
  signal(SIGTERM, term_handler);
  sigaddset(&mask, SIGTERM);
#endif

#ifdef SIGINT
  signal(SIGINT, term_handler);
  sigaddset(&mask, SIGINT);
#endif

  sigprocmask(SIG_UNBLOCK, &mask, NULL);
}

static void uinput_close()
{
  if (ui_fd > 0) {
    ioctl(ui_fd, UI_DEV_DESTROY, NULL);
    close(ui_fd);
    ui_fd = -1;
  }
}

static int uinput_open(char *uinput_device_file_name)
{
  return open(uinput_device_file_name, O_RDWR);
}

static int uinput_setup(char *uinput_device_file_name)
{
  struct uinput_user_dev device;

  if(ui_fd > 0) {
    uinput_close();
  }

  ui_fd = uinput_open(uinput_device_file_name);
  if (ui_fd < 0) {
    perror("open");
    return -1;
  }

  memset( &device, 0, sizeof device );
  strcpy(device.name, name);
  device.id.bustype = 0;
  device.id.vendor  = 0;
  device.id.product = 0;
  device.id.version = 0;

  set_event_bits(ui_fd);

  if (write(ui_fd, &device, sizeof(device)) < sizeof(device)) {
    perror("write");
    close(ui_fd);
    ui_fd = -1;
    return -1;
  }

  ioctl(ui_fd, UI_DEV_CREATE, NULL);

  return ui_fd;
}

void send_event(int type, int code, int value)
{
  struct input_event event;

  if (ui_fd < 0) {
    return;
  }

  event.type = type;
  event.code = code;
  event.value = value;
  gettimeofday(&event.time, NULL);
  if (write(ui_fd, &event, sizeof(event)) < sizeof(event)) {
    perror("send_event");
  }
}

void report_scroll(int dy)
{
  send_event(EV_REL, REL_WHEEL, -dy);
}

static void event_loop(int number_of_event_device_files, int *event_fds, void *user_data)
{
  struct input_event events[64];
  size_t read_size;
  int i, j;
  fd_set fds;
  struct timeval select_delay;
  int select_return_value;
  int max_fd = 0;

  for (i = 0; i < number_of_event_device_files; i++) {
    if (event_fds[i] > max_fd) {
      max_fd = event_fds[i];
    }
  }
  while (running > 0) {
    select_delay.tv_sec = 1;
    select_delay.tv_usec = 0;
    FD_ZERO(&fds);
    for (i = 0; i < number_of_event_device_files; i++) {
      FD_SET(event_fds[i], &fds);
    }
    select_return_value = select(max_fd + 1, &fds, NULL, NULL, &select_delay);
    if (select_return_value < 0) {
      /* perror("event_loop"); */
      continue;
    }
    if (select_return_value == 0) {
      continue;
    }
    for (i = 0; i < number_of_event_device_files; i++) {
      if (FD_ISSET(event_fds[i], &fds)) {
        read_size = read(event_fds[i], events, sizeof(events));
        for (j = 0; j < (int)(read_size / sizeof(struct input_event)); j++){
          handle_event(event_fds[i], events[j], user_data);
        }
      }
    }
  }
}

int main(int argc, char *argv[])
{
  char **event_device_file_names = (char **)malloc(sizeof(char *) * argc);
  char *uinput_device_file_name = "/dev/input/uinput";
  int i;
  int *event_fds = (int *)malloc(sizeof(int) * argc);
  int number_of_event_device_files = 0;
  void *user_data;

  if (event_device_file_names == NULL || event_fds == NULL) {
    perror("main");
    exit(EXIT_FAILURE);
  }
  
  install_sighandler();

  for (i = 1; i < argc; i++) {
    if (strcmp(argv[i], "-h") == 0 ||
        strcmp(argv[i], "--help") == 0) {
      print_usage(stdout);
      exit(EXIT_SUCCESS);
    } else if (strcmp(argv[i], "-v") == 0 ||
               strcmp(argv[i], "--version") == 0) {
      print_version();
      exit(EXIT_SUCCESS);
    } else if (strcmp(argv[i], "-u") == 0 ||
               strcmp(argv[i], "--uinput_device") == 0) {
      if (i + 1 >= argc) {
        print_usage(stderr);
        exit(EXIT_FAILURE);
      }
      i++;
      uinput_device_file_name = argv[i];
    } else {
      event_device_file_names[number_of_event_device_files] = argv[i];
      number_of_event_device_files++;
    }
  }

  if (number_of_event_device_files == 0) {
    print_usage(stderr);
    exit(EXIT_FAILURE);
  }

  running = uinput_setup(uinput_device_file_name);
  if (running < 0) {
    printf("uinput initialization failed. Make sure uinput module is loaded "
           "or available in the kernel.  Continue anyway.\n");
    running = 1;
  }

  for (i = 0; i < number_of_event_device_files; i++) {
    if (strcmp(event_device_file_names[i], "-") == 0) {
      event_fds[i] = STDIN_FILENO;
    } else {
      event_fds[i] = open(event_device_file_names[i], O_RDONLY);
    }
  }

  user_data = init_event_handler(number_of_event_device_files, event_fds);

  event_loop(number_of_event_device_files, event_fds, user_data);

  finalize_event_hander(user_data);
  for (i = 0; i < number_of_event_device_files; i++) {
    close(event_fds[i]);
  }

  uinput_close();

  exit(EXIT_SUCCESS);
}
